@@ -19,6 +19,9 @@ import numpy as np
 
 import scipy.constants as scc
 import scipy.signal as sig
+
+from scipy.optimize import fsolve
+
 import nds2
 
 '''Functions'''
@@ -1021,6 +1024,311 @@ def get_CSDs(channels,
     dataDict = data_CSDs(dataDict, averages, binwidth, overlap, averaging=averaging, makeTFs=makeTFs)
     return dataDict
 
+### get_all spectral density functions
+# returns every psd, csd used in calculating the average psd, csd via Welch's method
+def get_all_psds(x, fs=1.0, window='hann', nperseg=None, noverlap=None,
+                nfft=None, detrend='constant', return_onesided=True,
+                scaling='density', axis=-1):
+    '''Abuses the function scipy.signal.spectrogram() 
+    to return all power spectral densities which make up a PSD estimate.
+    Should be equivalent exactly to scipy.signal.spectrogram(mode='psd'),
+    but we want to use this method for get_all_csds() as well.
+    Included for consistency.
+
+    Parameters
+    ----------
+    x : array_like
+        Time series of measurement values
+    fs : float, optional
+        Sampling frequency of the `x` time series. Defaults to 1.0.
+    window : str or tuple or array_like, optional
+        Desired window to use. If `window` is a string or tuple, it is
+        passed to `get_window` to generate the window values, which are
+        DFT-even by default. See `get_window` for a list of windows and
+        required parameters. If `window` is array_like it will be used
+        directly as the window and its length must be nperseg.
+        Defaults to a Tukey window with shape parameter of 0.25.
+    nperseg : int, optional
+        Length of each segment. Defaults to None, but if window is str or
+        tuple, is set to 256, and if window is array_like, is set to the
+        length of the window.
+    noverlap : int, optional
+        Number of points to overlap between segments. If `None`,
+        ``noverlap = nperseg // 8``. Defaults to `None`.
+    nfft : int, optional
+        Length of the FFT used, if a zero padded FFT is desired. If
+        `None`, the FFT length is `nperseg`. Defaults to `None`.
+    detrend : str or function or `False`, optional
+        Specifies how to detrend each segment. If `detrend` is a
+        string, it is passed as the `type` argument to the `detrend`
+        function. If it is a function, it takes a segment and returns a
+        detrended segment. If `detrend` is `False`, no detrending is
+        done. Defaults to 'constant'.
+    return_onesided : bool, optional
+        If `True`, return a one-sided spectrum for real data. If
+        `False` return a two-sided spectrum. Defaults to `True`, but for
+        complex data, a two-sided spectrum is always returned.
+    scaling : { 'density', 'spectrum' }, optional
+        Selects between computing the power spectral density ('density')
+        where `Sxx` has units of V**2/Hz and computing the power
+        spectrum ('spectrum') where `Sxx` has units of V**2, if `x`
+        is measured in V and `fs` is measured in Hz. Defaults to
+        'density'.
+    axis : int, optional
+        Axis along which the spectrogram is computed; the default is over
+        the last axis (i.e. ``axis=-1``).
+    mode : str, optional
+        Defines what kind of return values are expected. Options are
+        ['psd', 'complex', 'magnitude', 'angle', 'phase']. 'complex' is
+        equivalent to the output of `stft` with no padding or boundary
+        extension. 'magnitude' returns the absolute magnitude of the
+        STFT. 'angle' and 'phase' return the complex angle of the STFT,
+        with and without unwrapping, respectively.
+
+    Returns
+    -------
+    f : ndarray
+        Array of sample frequencies.
+    t : ndarray
+        Array of segment times.
+    Sxxs : ndarray
+        Power spectral densities of x. By default, the last axis of Sxx corresponds
+        to the segment times.
+
+    '''
+    freqs, time, Fxx = sig.spectrogram(x, fs=fs, window=window, nperseg=nperseg, noverlap=noverlap,
+                                    nfft=nfft, detrend=detrend, return_onesided=return_onesided,
+                                    scaling=scaling, axis=axis, mode='complex')
+    Sxxs = np.abs( 2 * np.conjugate(Fxx) * Fxx )
+
+    return freqs, time, Sxxs
+
+def get_all_csds(x, y, fs=1.0, window='hann', nperseg=None, noverlap=None,
+                nfft=None, detrend='constant', return_onesided=True,
+                scaling='density', axis=-1):
+    '''Abuses the function scipy.signal.spectrogram() 
+    to return all cross spectral densities which make up a CSD estimate.
+
+    Parameters
+    ----------
+    x : array_like
+        Time series of measurement values
+    y : array_like
+        Time series of measurement values
+    fs : float, optional
+        Sampling frequency of the `x` time series. Defaults to 1.0.
+    window : str or tuple or array_like, optional
+        Desired window to use. If `window` is a string or tuple, it is
+        passed to `get_window` to generate the window values, which are
+        DFT-even by default. See `get_window` for a list of windows and
+        required parameters. If `window` is array_like it will be used
+        directly as the window and its length must be nperseg.
+        Defaults to a Tukey window with shape parameter of 0.25.
+    nperseg : int, optional
+        Length of each segment. Defaults to None, but if window is str or
+        tuple, is set to 256, and if window is array_like, is set to the
+        length of the window.
+    noverlap : int, optional
+        Number of points to overlap between segments. If `None`,
+        ``noverlap = nperseg // 8``. Defaults to `None`.
+    nfft : int, optional
+        Length of the FFT used, if a zero padded FFT is desired. If
+        `None`, the FFT length is `nperseg`. Defaults to `None`.
+    detrend : str or function or `False`, optional
+        Specifies how to detrend each segment. If `detrend` is a
+        string, it is passed as the `type` argument to the `detrend`
+        function. If it is a function, it takes a segment and returns a
+        detrended segment. If `detrend` is `False`, no detrending is
+        done. Defaults to 'constant'.
+    return_onesided : bool, optional
+        If `True`, return a one-sided spectrum for real data. If
+        `False` return a two-sided spectrum. Defaults to `True`, but for
+        complex data, a two-sided spectrum is always returned.
+    scaling : { 'density', 'spectrum' }, optional
+        Selects between computing the power spectral density ('density')
+        where `Sxx` has units of V**2/Hz and computing the power
+        spectrum ('spectrum') where `Sxx` has units of V**2, if `x`
+        is measured in V and `fs` is measured in Hz. Defaults to
+        'density'.
+    axis : int, optional
+        Axis along which the spectrogram is computed; the default is over
+        the last axis (i.e. ``axis=-1``).
+    mode : str, optional
+        Defines what kind of return values are expected. Options are
+        ['psd', 'complex', 'magnitude', 'angle', 'phase']. 'complex' is
+        equivalent to the output of `stft` with no padding or boundary
+        extension. 'magnitude' returns the absolute magnitude of the
+        STFT. 'angle' and 'phase' return the complex angle of the STFT,
+        with and without unwrapping, respectively.
+
+    Returns
+    -------
+    f : ndarray
+        Array of sample frequencies.
+    t : ndarray
+        Array of segment times.
+    Sxys : ndarray
+        Cross spectral densities of x and y. By default, the last axis of Sxys corresponds
+        to the segment times.
+
+    '''
+    freqs, time, Fxx = sig.spectrogram(x, fs=fs, window=window, nperseg=nperseg, noverlap=noverlap,
+                                    nfft=nfft, detrend=detrend, return_onesided=return_onesided,
+                                    scaling=scaling, axis=axis, mode='complex')
+    freqs, time, Fyy = sig.spectrogram(y, fs=fs, window=window, nperseg=nperseg, noverlap=noverlap,
+                                    nfft=nfft, detrend=detrend, return_onesided=return_onesided,
+                                    scaling=scaling, axis=axis, mode='complex')
+    Sxys = 2 * np.conjugate(Fxx) * Fyy 
+
+    return freqs, time, Sxys
+
+### Convenience functions for returning median-averaged psds and csds from get_all functions
+def get_median_psd(Saas):
+    '''Calculate the median PSD from a bunch of PSDs.  No bias is applied.
+    Example:
+    Sxxs = get_all_psds(x)
+    Sxx_med = get_median_psd(Sxxs)
+    '''
+    Saa_med = np.median(Saas, axis=1)
+    return Saa_med
+
+def get_median_csd(Sabs):
+    '''Calculate the median CSD from a bunch of CSDs.  No bias is applied.
+    Example:
+    Sxys = get_all_csds(x, y)
+    Sxy_med = get_median_csd(Sxys)
+    '''
+    Sab_med = np.median(np.real(Sabs), axis=1) + 1j * np.median(np.imag(Sabs), axis=1)
+    return Sab_med
+
+def get_coherence(Saa_med, Sbb_med, Sab_med):
+    '''Calculate the coherence from given PSDs and CSDs.
+    Example:
+    Sxxs = get_all_psds(x)
+    Syys = get_all_psds(y)
+    Sxys = get_all_csds(x, y)
+
+    Sxx_med = get_median_psd(Sxxs)
+    Syy_med = get_median_psd(Syys)
+    Sxy_med = get_median_csd(Sxys)
+
+    cohs_med = get_coherence(Sxx_med, Syy_med, Sxy_med)
+    '''
+    coherence_med = np.abs(Sab_med)**2 / (Saa_med * Sbb_med)
+    return coherence_med
+
+### Equivalent convenience functions for mean-averaged psds and csds.
+def get_mean_psd(Saas):
+    '''Calculate the mean PSD from a bunch of PSDs.  No bias is applied.
+    Example:
+    Sxxs = get_all_psds(x)
+    Sxx_mean = get_mean_psd(Sxxs)
+    '''
+    Saa_mean = np.mean(Saas, axis=1)
+    return Saa_mean
+
+def get_mean_csd(Sabs):
+    '''Calculate the mean CSD from a bunch of CSDs.  No bias is applied.
+    Example:
+    Sxys = get_all_csds(x, y)
+    Sxy_mean = get_mean_csd(Sxys)
+    '''
+    Sab_mean = np.mean(np.real(Sabs), axis=1) + 1j * np.mean(np.imag(Sabs), axis=1)
+    return Sab_mean
+
+### Mean-to-median bias functions for cross spectral densities
+# Outward-facing function is biases_from_coherences()
+def median_coherence_from_power_ratio(power_ratio):
+    '''Given a power_ratio of uncorrelated noise over correlated noise, 
+    power_ratio = sigma_uncorr^2 / sigma_corr^2,
+    returns the power coherence gamma^2 that would result from median-averaged PSDs and CSDs.
+    '''
+    med_coh =   (power_ratio**2 * np.log(1 + 1/np.sqrt(1 + power_ratio))**2) \
+                / (4 * np.log(2)**2 * (1 + power_ratio) * (2 + power_ratio - 2 * np.sqrt(1 + power_ratio)) )
+    return med_coh
+
+def residual_median_coherence(input_array, median_coherence_desired):
+    '''Finds the residual between the coherence desired and the numerically attempt from fsolve().
+    For use as the function input to fsolve() in bias_from_median_coherence().
+    Inputs:
+    input_array = array of length one, required by fsolve to find the power ratio numerically
+    median_coherence_desired = median coherence estimated from the CSD signals.
+    Output:
+    residual = array of length one with residual from median_coherence() function
+    '''
+    func_results = median_coherence_from_power_ratio(input_array[0])
+    residual = [func_results - median_coherence_desired]
+    return residual
+
+def bias_from_median_coherence_and_power_ratio(median_coherence, power_ratio):
+    '''Calculates the mean-to-median bias factor from the median_coherence = gamma^2
+    and the power_ratio = epsilon.
+    '''
+    bias = np.log(2) * np.sqrt( (1 + power_ratio) * median_coherence )
+    return bias
+
+def bias_from_median_coherence(median_coherence_estimated, initial_power_ratio=0.1):
+    '''Estimates the median/mean bias = b given some median_coherence = gamma^2.
+
+    Numerically solves for the uncorrelated/correated power_ratio = epsilon,
+    using the median_coherence.
+    Uses scipy.optimize.fsolve() to find the root of residual_median_coherence().
+    fsolve() seems to work for gamma^2 between (0.999 and 1e-6)
+
+    This function requires bias_from_median_coherence_and_power_ratio(),
+    residual_median_coherence(), median_coherence(), and fsolve() functions.
+
+    Inputs:
+    median_coherence_estimated  =   median coherence estimated from |<x,z>|^2/(<x,x> <z,z>) 
+                                    where all spectral densities are median-averaged
+    initial_power_ratio         =   initial guess of the power_ratio, default is 1.0
+    Output:
+    bias    =   median/mean bias factor.  Divide median-averaged cross spectral density <x,z> 
+                by bias to recover the mean-avearged cross spectral density.             
+    '''
+    # Numerically estimate the power ratio epsilon from the median coherence
+    fsolve_array = fsolve(  residual_median_coherence, 
+                            [initial_power_ratio], 
+                            args=(median_coherence_estimated))
+    power_ratio = fsolve_array[0]
+
+    # Find the bias factor
+    bias = bias_from_median_coherence_and_power_ratio(  median_coherence_estimated, 
+                                                        power_ratio)
+    return bias
+
+def biases_from_median_coherences(median_coherences, initial_power_ratios=None):
+    '''Returns array of CSD mean-to-median biases from an array of median_coherences.
+    Inputs:
+    median_coherences =    median-averaged coherences 
+    initial_power_ratios =  power ratios to start estimation with.  
+                            If None, uses 0.1 for all.  Default is None.
+    Outputs:
+    biases =    array of biases calculated from the coherences
+
+    Example:
+    Sxxs = get_all_psds(x)
+    Syys = get_all_psds(y)
+    Sxys = get_all_csds(x, y)
+
+    Sxx_med = get_median_psd(Sxxs)
+    Syy_med = get_median_psd(Syys)
+    Sxy_med = get_median_csd(Sxys)
+
+    cohs_med = get_coherence(Sxx_med, Syy_med, Sxy_med)
+    biases = biases_from_coherences(cohs_med)
+
+    Sxy_med_bias_corrected = Sxy_med / biases
+    '''
+    if initial_power_ratios is None:
+        initial_power_ratios = 0.1 * np.ones_like(median_coherences)
+    biases = np.array([])
+    for coherence, initial_power_ratio in zip(median_coherences, initial_power_ratios):
+        bias = bias_from_median_coherence(coherence, initial_power_ratio=initial_power_ratio)
+        biases = np.append(biases, bias)
+    return biases
+
+
 ### Modify data based on thresholds
 def coherence_threshold_applier(coherenceVector, coherenceThreshold=0.9):
     '''Apply a coherence threshold to a vector, returns all indicies of the vector above the threshold'''
